@@ -179,12 +179,12 @@ clusterBanksy <-
 
 
 #' @importFrom leidenAlg leiden.community
-#' @importFrom igraph as.undirected cluster_louvain
+#' @importFrom igraph as_undirected cluster_louvain
 runGraphBased <- function(cluster_matrices, algo, 
                          k_neighbors, resolution, leiden.iter, seed) {
     lapply(cluster_matrices, function(cmat) {
         lapply(k_neighbors, function(k) {
-            graph <- as.undirected(getGraph(cmat, k))
+            graph <- igraph::as_undirected(getGraph(cmat, k))
             lapply(resolution, function(res){
                 verbose.seed(seed)
                 memb <- switch(
@@ -550,7 +550,7 @@ smoother <-
                 neighbor_props <- table(neighbor_labels) /
                     length(neighbor_labels)
                 # Change label based on condition
-                if (any(neighbor_props > 0.5)) {
+                if (any(neighbor_props > prop_thres)) {
                     labels_update[i] <- as.numeric(
                         names(which.max(neighbor_props))
                     )
