@@ -51,6 +51,20 @@
 #' }
 #' @param center A logical scalar specifying whether to center higher order
 #'   harmonics in local neighborhoods.
+#' @param chunk_size A integer scalar specifying the number of rows / genes of 
+#'   the neighborhood cell matrix to compute. Must be strictly less than floor 
+#'   of 2e31-1 / number of cells, though this may still give rise to negative
+#'   length vector errors. For safety, set this to less than 
+#'   2e31-1 * row_limit_factor / number of cells. This is automatically 
+#'   computed based on the latter, but can be specified.
+#' @param parallel A logical scalar specifying whether to compute chunks in 
+#'   parallel using bplapply. Not implemented for Windows.
+#' @param num_cores A integer scalar specifying the number of cores to use 
+#'   if parallel is TRUE.
+#' @param row_limit_factor A numeric scalar specifying the safety factor 
+#'   applied to the maximum vector length (2^31-1) when computing chunk sizes. 
+#'   Accounts for data.table overhead during grouping operations. 
+#'   Default is 0.75.
 #' @param verbose A logical scalar specifying verbosity. 
 #'
 #' @importFrom SummarizedExperiment assay assay<- assayNames
@@ -83,6 +97,10 @@ computeBanksy <- function(se,
                           seed = NULL,
                           dimensions = "all",
                           center = TRUE,
+                          chunk_size = NULL,
+                          parallel = FALSE,
+                          num_cores = NULL,
+                          row_limit_factor = 0.75,
                           verbose = TRUE) {
     
     # Check args
@@ -119,9 +137,12 @@ computeBanksy <- function(se,
     })
 
     # Compute harmonics with different k_geoms
-    center <- c(FALSE, rep(TRUE, length(M) - 1))
+    center <- c(FALSE, rep(center, length(M) - 1))
     har <- Map(function(knn_df, M, center) {
-        out <- computeHarmonics(expr, knn_df, M, center, verbose = verbose)
+        out <- computeHarmonics(expr, knn_df, M, center, 
+                                verbose = verbose, chunk_size = chunk_size, 
+                                parallel = parallel, num_cores = num_cores, 
+                                row_limit_factor = row_limit_factor)
         rownames(out) <- rownames(expr)
         out
     }, knn_list, M, center)
@@ -288,11 +309,11 @@ getBanksyMatrix <- function(se,
     joint
 }
 
-#' @importFrom matrixStats rowSds
+#' @importFrom MatrixGenerics rowSds
 scaler <- function(x) {
     x <- as.matrix(x)
     rm <- rowMeans(x)
-    rsd <- rowSds(x)
+    rsd <- MatrixGenerics::rowSds(x)
     x <- (x - rm) / rsd
     x[is.nan(x)] <- 0
     return(x)
@@ -371,29 +392,116 @@ computeNeighbors <- function(locs,
 }
 
 
-computeHarmonics <- function(gcm, knn_df, M, center, verbose) {
+computeHarmonics <- function(gcm, knn_df, M, center, verbose, chunk_size = NULL, parallel = FALSE, num_cores = NULL, row_limit_factor = 0.75) {
     from <- to <- weight <- phi <- .N <- count <- . <- NULL
     j <- sqrt(as.complex(-1))
-
     mean_k <- round(mean(knn_df[, .(count = .N), by = from]$count), 1)
-
-    if (verbose) message("Computing harmonic m = ", M)
-    if (verbose) message("Using ", mean_k, " neighbors")
-    if (center) {
-        if (verbose) message("Centering")
-        aggr <- knn_df[, abs(
-            fscale(gcm[, to, drop = FALSE]) %*% (weight * exp(j * M * phi))
-        ), by = from]
+    
+    total_rows <- as.double(nrow(gcm)) * ncol(gcm)
+    max_rows <- (2^31 - 1) * row_limit_factor
+    if (total_rows > max_rows || !is.null(chunk_size)) {
+        if (verbose) message("Computing neighborhood matrices in chunks...")
+        # Automatically compute max chunk size
+        max_chunk_size <- floor(max_rows / ncol(gcm))
+        if (!is.null(chunk_size)) {
+            # If user specifies a chunk size
+            if (chunk_size > max_chunk_size) {
+                stop('Specified chunk_size too large. Must be smaller than ', floor(max_rows / ncol(gcm)))
+            }
+            max_chunk_size <- chunk_size
+        }
     } else {
-        aggr <- knn_df[, abs(
-            gcm[, to, drop = FALSE] %*% (weight * exp(j * M * phi))
-        ), by = from]
+        # Process whole dataset, backward compatible
+        max_chunk_size <- nrow(gcm)
     }
-    ncm <- matrix(aggr$V1, nrow = nrow(gcm), ncol = ncol(gcm))
+    
+    num_chunks <- ceiling(nrow(gcm) / max_chunk_size)
+    if (verbose && num_chunks > 1) {
+        message("Processing in ", num_chunks, " chunks of max ", max_chunk_size, " genes each")
+        if (parallel) {
+            if (.Platform$OS.type == "windows") {
+                message("Parallel processing not supported on Windows - using sequential processing")
+            } else {
+                if (!is.null(num_cores)) {
+                    message("Using parallel processing with ", num_cores, " cores")
+                } else {
+                    message("Using parallel processing with default backend")
+                }
+            }
+        }
+    }
+    
+    # Conditionally set the apply function and backend
+    if (parallel) {
+        if (.Platform$OS.type == "windows") {
+            if (verbose) message("Parallel processing not supported on Windows. Using sequential processing.")
+            apply_fun <- lapply
+        } else {
+            if (!requireNamespace("BiocParallel", quietly = TRUE)) {
+                stop("BiocParallel package is required for parallel processing. Please install it with: BiocManager::install('BiocParallel')")
+            }
+            
+            if (!is.null(num_cores)) {
+                bp_param <- BiocParallel::MulticoreParam(workers = num_cores)
+                apply_fun <- function(x, fun) BiocParallel::bplapply(x, fun, BPPARAM = bp_param)
+            } else {
+                apply_fun <- BiocParallel::bplapply
+            }
+        }
+    } else {
+        apply_fun <- lapply
+    }
+    
+    # Define chunk processing function
+    process_chunk <- function(chunk) {
+        start_idx <- (chunk - 1) * max_chunk_size + 1
+        end_idx <- min(chunk * max_chunk_size, nrow(gcm))
+        
+        if (verbose && num_chunks > 1 && (!parallel || .Platform$OS.type == "windows")) {
+            message("Processing chunk ", chunk, "/", num_chunks, " (rows ", start_idx, " to ", end_idx, ")")
+        }
+        
+        gcm_chunk <- gcm[start_idx:end_idx, , drop = FALSE]
+        
+        if (center) {
+            if (verbose && chunk == 1 && (!parallel || .Platform$OS.type == "windows")) message("Centering")
+            chunk_aggr <- knn_df[, abs(
+                fscale(gcm_chunk[, to, drop = FALSE]) %*% (weight * exp(j * M * phi))
+            ), by = from]
+        } else {
+            chunk_aggr <- knn_df[, abs(
+                gcm_chunk[, to, drop = FALSE] %*% (weight * exp(j * M * phi))
+            ), by = from]
+        }
+        
+        result_vector <- chunk_aggr$V1
+        rm(gcm_chunk, chunk_aggr)
+        
+        return(result_vector)
+    }
+    
+    # Process chunks
+    chunk_results <- apply_fun(1:num_chunks, process_chunk)
+    
+    # Initialize result matrix
+    ncm <- matrix(0, nrow = nrow(gcm), ncol = ncol(gcm))
+    
+    # Combine chunk results into final matrix
+    for (i in seq_along(chunk_results)) {
+        start_idx <- (i - 1) * max_chunk_size + 1
+        end_idx <- min(i * max_chunk_size, nrow(gcm))
+        ncm[start_idx:end_idx, ] <- matrix(chunk_results[[i]],
+                                           nrow = end_idx - start_idx + 1,
+                                           ncol = ncol(gcm))
+    }
+    
+    # Clean up chunk results and run garbage collection
+    rm(chunk_results)
+
     rownames(ncm) <- rownames(gcm)
     colnames(ncm) <- colnames(gcm)
-    if (verbose) message("Done")
 
+    if (verbose) message("Done")
     return(ncm)
 }
 
